@@ -231,23 +231,6 @@ soup_expectation_get_type (void)
 }
 
 /* enumerations from "soup-misc.h" */
-GType
-soup_ssl_error_get_type (void)
-{
-	static GType etype = 0;
-	if (G_UNLIKELY (etype == 0)) {
-		static const GEnumValue values[] = {
-			{ SOUP_SSL_ERROR_HANDSHAKE_NEEDS_READ, "SOUP_SSL_ERROR_HANDSHAKE_NEEDS_READ", "handshake-needs-read" },
-			{ SOUP_SSL_ERROR_HANDSHAKE_NEEDS_WRITE, "SOUP_SSL_ERROR_HANDSHAKE_NEEDS_WRITE", "handshake-needs-write" },
-			{ SOUP_SSL_ERROR_CERTIFICATE, "SOUP_SSL_ERROR_CERTIFICATE", "certificate" },
-			{ SOUP_SSL_ERROR_HANDSHAKE_FAILED, "SOUP_SSL_ERROR_HANDSHAKE_FAILED", "handshake-failed" },
-			{ 0, NULL, NULL }
-		};
-		etype = g_enum_register_static (g_intern_static_string ("SoupSSLError"), values);
-	}
-	return etype;
-}
-
 GType
 soup_connection_state_get_type (void)
 {
