@@ -51,8 +51,6 @@ GType soup_expectation_get_type (void) G_GNUC_CONST;
 #define SOUP_TYPE_EXPECTATION (soup_expectation_get_type ())
 
 /* enumerations from "soup-misc.h" */
-GType soup_ssl_error_get_type (void) G_GNUC_CONST;
-#define SOUP_TYPE_SSL_ERROR (soup_ssl_error_get_type ())
 GType soup_connection_state_get_type (void) G_GNUC_CONST;
 #define SOUP_TYPE_CONNECTION_STATE (soup_connection_state_get_type ())
 
