@@ -29,11 +29,7 @@
 #include "soup-http-input-stream.h"
 #include "soup-session.h"
 
-static void soup_http_input_stream_seekable_iface_init (GSeekableIface *seekable_iface);
-
-G_DEFINE_TYPE_WITH_CODE (SoupHTTPInputStream, soup_http_input_stream, G_TYPE_INPUT_STREAM,
-			 G_IMPLEMENT_INTERFACE (G_TYPE_SEEKABLE,
-						soup_http_input_stream_seekable_iface_init))
+G_DEFINE_TYPE (SoupHTTPInputStream, soup_http_input_stream, G_TYPE_INPUT_STREAM)
 
 typedef void (*SoupHTTPInputStreamCallback)(GInputStream *);
 
@@ -51,8 +47,7 @@ typedef struct {
 	SoupHTTPInputStreamCallback finished_cb;
 	SoupHTTPInputStreamCallback cancelled_cb;
 
-	guchar *leftover_buffer;
-	gsize leftover_bufsize, leftover_offset;
+	GQueue *leftover_queue;
 
 	guchar *caller_buffer;
 	gsize caller_bufsize, caller_nread;
@@ -89,21 +84,6 @@ static gboolean soup_http_input_stream_close_finish (GInputStream         *strea
 						     GAsyncResult         *result,
 						     GError              **error);
 
-static goffset  soup_http_input_stream_tell (GSeekable            *seekable);
-
-static gboolean soup_http_input_stream_can_seek (GSeekable            *seekable);
-static gboolean soup_http_input_stream_seek (GSeekable            *seekable,
-					     goffset offset,
-					     GSeekType type,
-					     GCancellable         *cancellable,
-					     GError              **error);
-
-static gboolean soup_http_input_stream_can_truncate (GSeekable            *seekable);
-static gboolean soup_http_input_stream_truncate (GSeekable            *seekable,
-						 goffset offset,
-						 GCancellable         *cancellable,
-						 GError              **error);
-
 static void soup_http_input_stream_got_headers (SoupMessage *msg, gpointer stream);
 static void soup_http_input_stream_got_chunk (SoupMessage *msg, SoupBuffer *chunk, gpointer stream);
 static void soup_http_input_stream_finished (SoupMessage *msg, gpointer stream);
@@ -120,7 +100,9 @@ soup_http_input_stream_finalize (GObject *object)
 	g_signal_handlers_disconnect_by_func (priv->msg, G_CALLBACK (soup_http_input_stream_got_chunk), stream);
 	g_signal_handlers_disconnect_by_func (priv->msg, G_CALLBACK (soup_http_input_stream_finished), stream);
 	g_object_unref (priv->msg);
-	g_free (priv->leftover_buffer);
+
+	g_queue_foreach (priv->leftover_queue, (GFunc) soup_buffer_free, NULL);
+	g_queue_free (priv->leftover_queue);
 
 	if (G_OBJECT_CLASS (soup_http_input_stream_parent_class)->finalize)
 		(*G_OBJECT_CLASS (soup_http_input_stream_parent_class)->finalize)(object);
@@ -144,20 +126,12 @@ soup_http_input_stream_class_init (SoupHTTPInputStreamClass *klass)
 	stream_class->close_finish = soup_http_input_stream_close_finish;
 }
 
-static void
-soup_http_input_stream_seekable_iface_init (GSeekableIface *seekable_iface)
-{
-	seekable_iface->tell = soup_http_input_stream_tell;
-	seekable_iface->can_seek = soup_http_input_stream_can_seek;
-	seekable_iface->seek = soup_http_input_stream_seek;
-	seekable_iface->can_truncate = soup_http_input_stream_can_truncate;
-	seekable_iface->truncate_fn = soup_http_input_stream_truncate;
-}
-
 static void
 soup_http_input_stream_init (SoupHTTPInputStream *stream)
 {
-	;
+	SoupHTTPInputStreamPrivate *priv = SOUP_HTTP_INPUT_STREAM_GET_PRIVATE (stream);
+
+	priv->leftover_queue = g_queue_new ();
 }
 
 static void
@@ -198,7 +172,7 @@ soup_http_input_stream_queue_message (SoupHTTPInputStream *stream)
  *
  * Returns: a new #GInputStream.
  **/
-SoupHTTPInputStream *
+GInputStream *
 soup_http_input_stream_new (SoupSession *session, SoupMessage *msg)
 {
 	SoupHTTPInputStream *stream;
@@ -221,7 +195,7 @@ soup_http_input_stream_new (SoupSession *session, SoupMessage *msg)
 			  G_CALLBACK (soup_http_input_stream_finished), stream);
 
 	soup_http_input_stream_queue_message (stream);
-	return stream;
+	return (GInputStream *)stream;
 }
 
 static void
@@ -249,7 +223,7 @@ soup_http_input_stream_got_headers (SoupMessage *msg, gpointer stream)
 
 static void
 soup_http_input_stream_got_chunk (SoupMessage *msg, SoupBuffer *chunk_buffer,
-					 gpointer stream)
+				  gpointer stream)
 {
 	SoupHTTPInputStreamPrivate *priv = SOUP_HTTP_INPUT_STREAM_GET_PRIVATE (stream);
 	const gchar *chunk = chunk_buffer->data;
@@ -261,12 +235,8 @@ soup_http_input_stream_got_chunk (SoupMessage *msg, SoupBuffer *chunk_buffer,
 	if (!SOUP_STATUS_IS_SUCCESSFUL (msg->status_code))
 		return;
 
-	/* Sanity check */
-	if (priv->caller_bufsize == 0 || priv->leftover_bufsize != 0)
-		g_warning ("soup_http_input_stream_got_chunk called again before previous chunk was processed");
-
 	/* Copy what we can into priv->caller_buffer */
-	if (priv->caller_bufsize > priv->caller_nread) {
+	if (priv->caller_bufsize > priv->caller_nread && priv->leftover_queue->length == 0) {
 		gsize nread = MIN (chunk_size, priv->caller_bufsize - priv->caller_nread);
 
 		memcpy (priv->caller_buffer + priv->caller_nread, chunk, nread);
@@ -277,20 +247,13 @@ soup_http_input_stream_got_chunk (SoupMessage *msg, SoupBuffer *chunk_buffer,
 	}
 
 	if (chunk_size > 0) {
-		/* Copy the rest into priv->leftover_buffer. If
-		 * there's already some data there, realloc and
-		 * append. Otherwise just copy.
-		 */
-		if (priv->leftover_bufsize) {
-			priv->leftover_buffer = g_realloc (priv->leftover_buffer,
-							   priv->leftover_bufsize + chunk_size);
-			memcpy (priv->leftover_buffer + priv->leftover_bufsize,
-				chunk, chunk_size);
-			priv->leftover_bufsize += chunk_size;
+		if (priv->leftover_queue->length > 0) {
+			g_queue_push_tail (priv->leftover_queue, soup_buffer_copy (chunk_buffer));
 		} else {
-			priv->leftover_bufsize = chunk_size;
-			priv->leftover_buffer = g_memdup (chunk, chunk_size);
-			priv->leftover_offset = 0;
+			g_queue_push_head (priv->leftover_queue,
+					   soup_buffer_new_subbuffer (chunk_buffer,
+								      chunk_buffer->length - chunk_size,
+								      chunk_size));
 		}
 	}
 
@@ -385,19 +348,17 @@ read_from_leftover (SoupHTTPInputStreamPrivate *priv,
 		    gpointer buffer, gsize bufsize)
 {
 	gsize nread;
+	SoupBuffer *soup_buffer = (SoupBuffer *) g_queue_peek_head (priv->leftover_queue);
+	gboolean fits_in_buffer = soup_buffer->length <= bufsize;
 
-	if (priv->leftover_bufsize - priv->leftover_offset <= bufsize) {
-		nread = priv->leftover_bufsize - priv->leftover_offset;
-		memcpy (buffer, priv->leftover_buffer + priv->leftover_offset, nread);
+	nread = fits_in_buffer ? soup_buffer->length : bufsize;
+	memcpy (buffer, soup_buffer->data, nread);
 
-		g_free (priv->leftover_buffer);
-		priv->leftover_buffer = NULL;
-		priv->leftover_bufsize = priv->leftover_offset = 0;
-	} else {
-		nread = bufsize;
-		memcpy (buffer, priv->leftover_buffer + priv->leftover_offset, nread);
-		priv->leftover_offset += nread;
-	}
+	g_queue_pop_head (priv->leftover_queue);
+	if (!fits_in_buffer)
+		g_queue_push_head (priv->leftover_queue,
+				   soup_buffer_new_subbuffer (soup_buffer, nread, soup_buffer->length - nread));
+	soup_buffer_free (soup_buffer);
 
 	priv->offset += nread;
 	return nread;
@@ -431,16 +392,12 @@ static void
 send_sync_finished (GInputStream *stream)
 {
 	SoupHTTPInputStreamPrivate *priv = SOUP_HTTP_INPUT_STREAM_GET_PRIVATE (stream);
-	GError *error = NULL;
-
-	if (!g_cancellable_set_error_if_cancelled (priv->cancellable, &error))
-		set_error_if_http_failed (priv->msg, &error);
 
 	priv->got_headers_cb = NULL;
 	priv->finished_cb = NULL;
 
 	/* Wake up the main context iteration */
-	g_source_attach (g_idle_source_new (), NULL);
+	soup_add_completion (priv->async_context, NULL, NULL);
 }
 
 /**
@@ -489,13 +446,13 @@ soup_http_input_stream_read (GInputStream  *stream,
 {
 	SoupHTTPInputStreamPrivate *priv = SOUP_HTTP_INPUT_STREAM_GET_PRIVATE (stream);
 
+	/* If there is data leftover from a previous read, return it. */
+	if (priv->leftover_queue->length)
+		return read_from_leftover (priv, buffer, count);
+
 	if (priv->finished)
 		return 0;
 
-	/* If there is data leftover from a previous read, return it. */
-	if (priv->leftover_bufsize)
-		return read_from_leftover (priv, buffer, count);
-
 	/* No leftover data, accept one chunk from the network */
 	soup_http_input_stream_prepare_for_io (stream, cancellable, buffer, count);
 	while (!priv->finished && priv->caller_nread == 0 &&
@@ -521,8 +478,10 @@ soup_http_input_stream_close (GInputStream  *stream,
 {
 	SoupHTTPInputStreamPrivate *priv = SOUP_HTTP_INPUT_STREAM_GET_PRIVATE (stream);
 
-	if (!priv->finished)
+	if (!priv->finished) {
+		soup_session_unpause_message (priv->session, priv->msg);
 		soup_session_cancel_message (priv->session, priv->msg, SOUP_STATUS_CANCELLED);
+	}
 
 	return TRUE;
 }
@@ -536,44 +495,11 @@ wrapper_callback (GObject *source_object, GAsyncResult *res,
 
 	g_input_stream_clear_pending (stream);
 	if (priv->outstanding_callback)
-		(*priv->outstanding_callback)(source_object, res, user_data);
+		(*priv->outstanding_callback) (source_object, res, user_data);
 	priv->outstanding_callback = NULL;
 	g_object_unref (stream);
 }
 
-static void
-send_async_thread (GSimpleAsyncResult *res,
-		   GObject *object,
-		   GCancellable *cancellable)
-{
-	GError *error = NULL;
-	gboolean success;
-
-	success = soup_http_input_stream_send_internal (G_INPUT_STREAM (object),
-							cancellable, &error);
-	g_simple_async_result_set_op_res_gboolean (res, success);
-	if (error) {
-		g_simple_async_result_set_from_error (res, error);
-		g_error_free (error);
-	}
-}
-
-static void
-soup_http_input_stream_send_async_in_thread (GInputStream        *stream,
-					     int                  io_priority,
-					     GCancellable        *cancellable,
-					     GAsyncReadyCallback  callback,
-					     gpointer             user_data)
-{
-	GSimpleAsyncResult *res;
-
-	res = g_simple_async_result_new (G_OBJECT (stream), callback, user_data,
-					 soup_http_input_stream_send_async_in_thread);
-	g_simple_async_result_run_in_thread (res, send_async_thread,
-					     io_priority, cancellable);
-	g_object_unref (res);
-}
-
 static void
 send_async_finished (GInputStream *stream)
 {
@@ -592,10 +518,8 @@ send_async_finished (GInputStream *stream)
 	priv->result = NULL;
 
 	g_simple_async_result_set_op_res_gboolean (result, error == NULL);
-	if (error) {
-		g_simple_async_result_set_from_error (result, error);
-		g_error_free (error);
-	}
+	if (error)
+		g_simple_async_result_take_error (result, error);
 	g_simple_async_result_complete (result);
 	g_object_unref (result);
 }
@@ -609,19 +533,11 @@ soup_http_input_stream_send_async_internal (GInputStream        *stream,
 {
 	SoupHTTPInputStreamPrivate *priv = SOUP_HTTP_INPUT_STREAM_GET_PRIVATE (stream);
 
+	g_return_if_fail (priv->async_context == g_main_context_get_thread_default ());
+
 	g_object_ref (stream);
 	priv->outstanding_callback = callback;
 
-	/* If the session uses the default GMainContext, then we can do
-	 * async I/O directly. But if it has its own main context, it's
-	 * easier to just run it in another thread.
-	 */
-	if (soup_session_get_async_context (priv->session)) {
-		soup_http_input_stream_send_async_in_thread (stream, io_priority, cancellable,
-							     wrapper_callback, user_data);
-		return;
-	}
-
 	priv->got_headers_cb = send_async_finished;
 	priv->finished_cb = send_async_finished;
 
@@ -657,11 +573,10 @@ soup_http_input_stream_send_async (SoupHTTPInputStream *httpstream,
 	g_return_if_fail (SOUP_IS_HTTP_INPUT_STREAM (httpstream));
 
 	if (!g_input_stream_set_pending (istream, &error)) {
-		g_simple_async_report_gerror_in_idle (G_OBJECT (httpstream),
-						      callback,
-						      user_data,
-						      error);
-		g_error_free (error);
+		g_simple_async_report_take_gerror_in_idle (G_OBJECT (httpstream),
+							   callback,
+							   user_data,
+							   error);
 		return;
 	}
 	soup_http_input_stream_send_async_internal (istream, io_priority, cancellable,
@@ -709,10 +624,9 @@ read_async_done (GInputStream *stream)
 	priv->result = NULL;
 
 	if (g_cancellable_set_error_if_cancelled (priv->cancellable, &error) ||
-	    set_error_if_http_failed (priv->msg, &error)) {
-		g_simple_async_result_set_from_error (result, error);
-		g_error_free (error);
-	} else
+	    set_error_if_http_failed (priv->msg, &error))
+		g_simple_async_result_take_error (result, error);
+	else
 		g_simple_async_result_set_op_res_gssize (result, priv->caller_nread);
 
 	priv->got_chunk_cb = NULL;
@@ -736,31 +650,22 @@ soup_http_input_stream_read_async (GInputStream        *stream,
 	SoupHTTPInputStreamPrivate *priv = SOUP_HTTP_INPUT_STREAM_GET_PRIVATE (stream);
 	GSimpleAsyncResult *result;
 
-	/* If the session uses the default GMainContext, then we can do
-	 * async I/O directly. But if it has its own main context, we fall
-	 * back to the async-via-sync-in-another-thread implementation.
-	 */
-	if (soup_session_get_async_context (priv->session)) {
-		G_INPUT_STREAM_CLASS (soup_http_input_stream_parent_class)->
-		read_async (stream, buffer, count, io_priority,
-			    cancellable, callback, user_data);
-		return;
-	}
+	g_return_if_fail (priv->async_context == g_main_context_get_thread_default ());
 
 	result = g_simple_async_result_new (G_OBJECT (stream),
 					    callback, user_data,
 					    soup_http_input_stream_read_async);
 
-	if (priv->finished) {
-		g_simple_async_result_set_op_res_gssize (result, 0);
+	if (priv->leftover_queue->length) {
+		gsize nread = read_from_leftover (priv, buffer, count);
+		g_simple_async_result_set_op_res_gssize (result, nread);
 		g_simple_async_result_complete_in_idle (result);
 		g_object_unref (result);
 		return;
 	}
 
-	if (priv->leftover_bufsize) {
-		gsize nread = read_from_leftover (priv, buffer, count);
-		g_simple_async_result_set_op_res_gssize (result, nread);
+	if (priv->finished) {
+		g_simple_async_result_set_op_res_gssize (result, 0);
 		g_simple_async_result_complete_in_idle (result);
 		g_object_unref (result);
 		return;
@@ -804,10 +709,8 @@ soup_http_input_stream_close_async (GInputStream        *stream,
 					    soup_http_input_stream_close_async);
 	success = soup_http_input_stream_close (stream, cancellable, &error);
 	g_simple_async_result_set_op_res_gboolean (result, success);
-	if (error) {
-		g_simple_async_result_set_from_error (result, error);
-		g_error_free (error);
-	}
+	if (error)
+		g_simple_async_result_take_error (result, error);
 
 	g_simple_async_result_complete_in_idle (result);
 	g_object_unref (result);
@@ -822,97 +725,6 @@ soup_http_input_stream_close_finish (GInputStream  *stream,
 	return TRUE;
 }
 
-static goffset
-soup_http_input_stream_tell (GSeekable *seekable)
-{
-	SoupHTTPInputStreamPrivate *priv = SOUP_HTTP_INPUT_STREAM_GET_PRIVATE (seekable);
-
-	return priv->offset;
-}
-
-static gboolean
-soup_http_input_stream_can_seek (GSeekable *seekable)
-{
-	return TRUE;
-}
-
-extern void soup_message_io_cleanup (SoupMessage *msg);
-
-static gboolean
-soup_http_input_stream_seek (GSeekable     *seekable,
-			     goffset        offset,
-			     GSeekType      type,
-			     GCancellable  *cancellable,
-			     GError       **error)
-{
-	GInputStream *stream = G_INPUT_STREAM (seekable);
-	SoupHTTPInputStreamPrivate *priv = SOUP_HTTP_INPUT_STREAM_GET_PRIVATE (seekable);
-	char *range;
-
-	if (type == G_SEEK_END) {
-		/* FIXME: we could send "bytes=-offset", but unless we
-		 * know the Content-Length, we wouldn't be able to
-		 * answer a tell() properly. We could find the
-		 * Content-Length by doing a HEAD...
-		 */
-
-		g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
-				     "G_SEEK_END not currently supported");
-		return FALSE;
-	}
-
-	if (!g_input_stream_set_pending (stream, error))
-		return FALSE;
-
-	soup_session_cancel_message (priv->session, priv->msg, SOUP_STATUS_CANCELLED);
-	soup_message_io_cleanup (priv->msg);
-
-	switch (type) {
-		case G_SEEK_CUR:
-			offset += priv->offset;
-		/* fall through */
-
-		case G_SEEK_SET:
-			range = g_strdup_printf ("bytes=%" G_GUINT64_FORMAT "-", (guint64)offset);
-			priv->offset = offset;
-			break;
-
-		case G_SEEK_END:
-			range = NULL; /* keep compilers happy */
-			g_return_val_if_reached (FALSE);
-			break;
-
-		default:
-			g_return_val_if_reached (FALSE);
-	}
-
-	soup_message_headers_remove (priv->msg->request_headers, "Range");
-	soup_message_headers_append (priv->msg->request_headers, "Range", range);
-	g_free (range);
-
-	soup_http_input_stream_queue_message (SOUP_HTTP_INPUT_STREAM (stream));
-
-	g_input_stream_clear_pending (stream);
-	return TRUE;
-}
-
-static gboolean
-soup_http_input_stream_can_truncate (GSeekable *seekable)
-{
-	return FALSE;
-}
-
-static gboolean
-soup_http_input_stream_truncate (GSeekable     *seekable,
-				 goffset        offset,
-				 GCancellable  *cancellable,
-				 GError       **error)
-{
-	g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
-			     "Truncate not allowed on input stream");
-	return FALSE;
-}
-
 SoupMessage *
 soup_http_input_stream_get_message (SoupHTTPInputStream *httpstream)
 {
