@@ -6,9 +6,9 @@
 #ifndef SOUP_PASSWORD_MANAGER_GNOME_H
 #define SOUP_PASSWORD_MANAGER_GNOME_H 1
 
-#include <libsoup/soup-password-manager.h>
+#include "soup-password-manager.h"
+#include "soup-gnome-features.h"
 
-#define SOUP_TYPE_PASSWORD_MANAGER_GNOME            (soup_password_manager_gnome_get_type ())
 #define SOUP_PASSWORD_MANAGER_GNOME(object)         (G_TYPE_CHECK_INSTANCE_CAST ((object), SOUP_TYPE_PASSWORD_MANAGER_GNOME, SoupPasswordManagerGNOME))
 #define SOUP_PASSWORD_MANAGER_GNOME_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), SOUP_TYPE_PASSWORD_MANAGER_GNOME, SoupPasswordManagerGNOMEClass))
 #define SOUP_IS_PASSWORD_MANAGER_GNOME(object)      (G_TYPE_CHECK_INSTANCE_TYPE ((object), SOUP_TYPE_PASSWORD_MANAGER_GNOME))
@@ -25,8 +25,4 @@ typedef struct {
 
 } SoupPasswordManagerGNOMEClass;
 
-GType soup_password_manager_gnome_get_type (void);
-
-SoupPasswordManager *soup_password_manager_gnome_new (void);
-
 #endif /* SOUP_PASSWORD_MANAGER_GNOME_H */
