@@ -101,8 +101,11 @@ typedef struct {
 	GMutex *host_lock;
 
 	GMainContext *async_context;
+	gboolean use_thread_context;
 
 	GResolver *resolver;
+
+	char **http_aliases, **https_aliases;
 } SoupSessionPrivate;
 #define SOUP_SESSION_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_SESSION, SoupSessionPrivate))
 
@@ -154,6 +157,7 @@ enum {
 	PROP_TLS_DATABASE,
 	PROP_SSL_STRICT,
 	PROP_ASYNC_CONTEXT,
+	PROP_USE_THREAD_CONTEXT,
 	PROP_TIMEOUT,
 	PROP_USER_AGENT,
 	PROP_ACCEPT_LANGUAGE,
@@ -162,6 +166,8 @@ enum {
 	PROP_ADD_FEATURE,
 	PROP_ADD_FEATURE_BY_TYPE,
 	PROP_REMOVE_FEATURE_BY_TYPE,
+	PROP_HTTP_ALIASES,
+	PROP_HTTPS_ALIASES,
 
 	LAST_PROP
 };
@@ -206,6 +212,10 @@ soup_session_init (SoupSession *session)
 	priv->resolver = g_resolver_get_default ();
 
 	priv->ssl_strict = TRUE;
+
+	priv->http_aliases = g_new (char *, 2);
+	priv->http_aliases[0] = (char *)g_intern_string ("*");
+	priv->http_aliases[1] = NULL;
 }
 
 static void
@@ -248,6 +258,9 @@ finalize (GObject *object)
 
 	g_object_unref (priv->resolver);
 
+	g_free (priv->http_aliases);
+	g_free (priv->https_aliases);
+
 	G_OBJECT_CLASS (soup_session_parent_class)->finalize (object);
 }
 
@@ -642,13 +655,40 @@ soup_session_class_init (SoupSessionClass *session_class)
 	 *
 	 * Alias for the #SoupSession:async-context property. (The
 	 * session's #GMainContext.)
-	 **/
+	 */
 	g_object_class_install_property (
 		object_class, PROP_ASYNC_CONTEXT,
 		g_param_spec_pointer (SOUP_SESSION_ASYNC_CONTEXT,
 				      "Async GMainContext",
 				      "The GMainContext to dispatch async I/O in",
 				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	/**
+	 * SOUP_SESSION_USE_THREAD_CONTEXT:
+	 *
+	 * Alias for the #SoupSession:use-thread-context property, qv.
+	 *
+	 * Since: 2.38
+	 */
+	/**
+	 * SoupSession:use-thread-context:
+	 *
+	 * If set, asynchronous operations in this session will run in
+	 * whatever the thread-default #GMainContext is at the time
+	 * they are started, rather than always occurring in a context
+	 * fixed at the session's construction time. "Bookkeeping"
+	 * tasks (like expiring idle connections) will happen in the
+	 * context that was thread-default at the time the session was
+	 * created.
+	 *
+	 * Since: 2.38
+	 */
+	g_object_class_install_property (
+		object_class, PROP_USE_THREAD_CONTEXT,
+		g_param_spec_boolean (SOUP_SESSION_USE_THREAD_CONTEXT,
+				      "Use thread-default GMainContext",
+				      "Whether to use thread-default main contexts",
+				      FALSE,
+				      G_PARAM_READWRITE));
 	/**
 	 * SOUP_SESSION_TIMEOUT:
 	 *
@@ -826,6 +866,69 @@ soup_session_class_init (SoupSessionClass *session_class)
 				    "Remove features of the given type from the session",
 				    SOUP_TYPE_SESSION_FEATURE,
 				    G_PARAM_READWRITE));
+	/**
+	 * SoupSession:http-aliases:
+	 *
+	 * A %NULL-terminated array of URI schemes that should be
+	 * considered to be aliases for "http". Eg, if this included
+	 * <literal>"dav"</literal>, than a URI of
+	 * <literal>dav://example.com/path</literal> would be treated
+	 * identically to <literal>http://example.com/path</literal>.
+	 * If the value is %NULL, then only "http" is recognized as
+	 * meaning "http".
+	 *
+	 * For backward-compatibility reasons, the default value for
+	 * this property is an array containing the single element
+	 * <literal>"*"</literal>, a special value which means that
+	 * any scheme except "https" is considered to be an alias for
+	 * "http".
+	 *
+	 * See also #SoupSession:https-aliases.
+	 *
+	 * Since: 2.38
+	 */
+	/**
+	 * SOUP_SESSION_HTTP_ALIASES:
+	 *
+	 * Alias for the #SoupSession:http-aliases property. (URI
+	 * schemes that will be considered aliases for "http".)
+	 *
+	 * Since: 2.38
+	 */
+	g_object_class_install_property (
+		object_class, PROP_HTTP_ALIASES,
+		g_param_spec_boxed (SOUP_SESSION_HTTP_ALIASES,
+				    "http aliases",
+				    "URI schemes that are considered aliases for 'http'",
+				    G_TYPE_STRV,
+				    G_PARAM_READWRITE));
+	/**
+	 * SoupSession:https-aliases:
+	 *
+	 * A comma-delimited list of URI schemes that should be
+	 * considered to be aliases for "https". See
+	 * #SoupSession:http-aliases for more information.
+	 *
+	 * The default value is %NULL, meaning that no URI schemes
+	 * are considered aliases for "https".
+	 *
+	 * Since: 2.38
+	 */
+	/**
+	 * SOUP_SESSION_HTTPS_ALIASES:
+	 *
+	 * Alias for the #SoupSession:https-aliases property. (URI
+	 * schemes that will be considered aliases for "https".)
+	 *
+	 * Since: 2.38
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_HTTPS_ALIASES,
+		g_param_spec_boxed (SOUP_SESSION_HTTPS_ALIASES,
+				    "https aliases",
+				    "URI schemes that are considered aliases for 'https'",
+				    G_TYPE_STRV,
+				    G_PARAM_READWRITE));
 }
 
 /* Converts a language in POSIX format and to be RFC2616 compliant    */
@@ -939,6 +1042,23 @@ load_ssl_ca_file (SoupSessionPrivate *priv)
 	g_error_free (error);
 }
 
+/* priv->http_aliases and priv->https_aliases are stored as arrays of
+ * *interned* strings, so we can't just use g_strdupv() to set them.
+ */
+static void
+set_aliases (char ***variable, char **value)
+{
+	int len = g_strv_length (value), i;
+
+	if (*variable)
+		g_free (*variable);
+
+	*variable = g_new (char *, len);
+	for (i = 0; i < len; i++)
+		(*variable)[i] = (char *)g_intern_string (value[i]);
+	(*variable)[i] = NULL;
+}
+
 static void
 set_property (GObject *object, guint prop_id,
 	      const GValue *value, GParamSpec *pspec)
@@ -1018,6 +1138,16 @@ set_property (GObject *object, guint prop_id,
 		if (priv->async_context)
 			g_main_context_ref (priv->async_context);
 		break;
+	case PROP_USE_THREAD_CONTEXT:
+		priv->use_thread_context = g_value_get_boolean (value);
+		if (priv->use_thread_context) {
+			if (priv->async_context)
+				g_main_context_unref (priv->async_context);
+			priv->async_context = g_main_context_get_thread_default ();
+			if (priv->async_context)
+				g_main_context_ref (priv->async_context);
+		}
+		break;
 	case PROP_TIMEOUT:
 		priv->io_timeout = g_value_get_uint (value);
 		break;
@@ -1064,6 +1194,12 @@ set_property (GObject *object, guint prop_id,
 	case PROP_REMOVE_FEATURE_BY_TYPE:
 		soup_session_remove_feature_by_type (session, g_value_get_gtype (value));
 		break;
+	case PROP_HTTP_ALIASES:
+		set_aliases (&priv->http_aliases, g_value_get_boxed (value));
+		break;
+	case PROP_HTTPS_ALIASES:
+		set_aliases (&priv->https_aliases, g_value_get_boxed (value));
+		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -1119,6 +1255,9 @@ get_property (GObject *object, guint prop_id,
 	case PROP_ASYNC_CONTEXT:
 		g_value_set_pointer (value, priv->async_context ? g_main_context_ref (priv->async_context) : NULL);
 		break;
+	case PROP_USE_THREAD_CONTEXT:
+		g_value_set_boolean (value, priv->use_thread_context);
+		break;
 	case PROP_TIMEOUT:
 		g_value_set_uint (value, priv->io_timeout);
 		break;
@@ -1134,12 +1273,60 @@ get_property (GObject *object, guint prop_id,
 	case PROP_IDLE_TIMEOUT:
 		g_value_set_uint (value, priv->idle_timeout);
 		break;
+	case PROP_HTTP_ALIASES:
+		g_value_set_boxed (value, priv->http_aliases);
+		break;
+	case PROP_HTTPS_ALIASES:
+		g_value_set_boxed (value, priv->https_aliases);
+		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
 	}
 }
 
+static gboolean
+uri_is_http (SoupSessionPrivate *priv, SoupURI *uri)
+{
+	int i;
+
+	if (uri->scheme == SOUP_URI_SCHEME_HTTP)
+		return TRUE;
+	else if (uri->scheme == SOUP_URI_SCHEME_HTTPS)
+		return FALSE;
+	else if (!priv->http_aliases)
+		return FALSE;
+
+	for (i = 0; priv->http_aliases[i]; i++) {
+		if (uri->scheme == priv->http_aliases[i])
+			return TRUE;
+	}
+
+	if (!priv->http_aliases[1] && !strcmp (priv->http_aliases[0], "*"))
+		return TRUE;
+	else
+		return FALSE;
+}
+
+static gboolean
+uri_is_https (SoupSessionPrivate *priv, SoupURI *uri)
+{
+	int i;
+
+	if (uri->scheme == SOUP_URI_SCHEME_HTTPS)
+		return TRUE;
+	else if (uri->scheme == SOUP_URI_SCHEME_HTTP)
+		return FALSE;
+	else if (!priv->https_aliases)
+		return FALSE;
+
+	for (i = 0; priv->https_aliases[i]; i++) {
+		if (uri->scheme == priv->https_aliases[i])
+			return TRUE;
+	}
+
+	return FALSE;
+}
 
 /**
  * soup_session_get_async_context:
@@ -1149,6 +1336,9 @@ get_property (GObject *object, guint prop_id,
  * context, so you will need to ref it yourself if you want it to
  * outlive its session.
  *
+ * If #SoupSession:use-thread-context is true, this will return the
+ * current thread-default main context.
+ *
  * Return value: (transfer none): @session's #GMainContext, which may
  * be %NULL
  **/
@@ -1160,7 +1350,10 @@ soup_session_get_async_context (SoupSession *session)
 	g_return_val_if_fail (SOUP_IS_SESSION (session), NULL);
 	priv = SOUP_SESSION_GET_PRIVATE (session);
 
-	return priv->async_context;
+	if (priv->use_thread_context)
+		return g_main_context_get_thread_default ();
+	else
+		return priv->async_context;
 }
 
 /* Hosts */
@@ -1247,31 +1440,126 @@ auth_manager_authenticate (SoupAuthManager *manager, SoupMessage *msg,
 				     method == SOUP_METHOD_OPTIONS || \
 				     method == SOUP_METHOD_PROPFIND)
 
-static void
-redirect_handler (SoupMessage *msg, gpointer user_data)
+#define SOUP_SESSION_WOULD_REDIRECT_AS_GET(session, msg) \
+	((msg)->status_code == SOUP_STATUS_SEE_OTHER || \
+	 ((msg)->status_code == SOUP_STATUS_FOUND && \
+	  !SOUP_METHOD_IS_SAFE ((msg)->method)) || \
+	 ((msg)->status_code == SOUP_STATUS_MOVED_PERMANENTLY && \
+	  (msg)->method == SOUP_METHOD_POST))
+
+#define SOUP_SESSION_WOULD_REDIRECT_AS_SAFE(session, msg) \
+	(((msg)->status_code == SOUP_STATUS_MOVED_PERMANENTLY || \
+	  (msg)->status_code == SOUP_STATUS_TEMPORARY_REDIRECT || \
+	  (msg)->status_code == SOUP_STATUS_FOUND) && \
+	 SOUP_METHOD_IS_SAFE ((msg)->method))
+
+static inline SoupURI *
+redirection_uri (SoupMessage *msg)
 {
-	SoupMessageQueueItem *item = user_data;
-	SoupSession *session = item->session;
 	const char *new_loc;
 	SoupURI *new_uri;
 
 	new_loc = soup_message_headers_get_one (msg->response_headers,
 						"Location");
-	g_return_if_fail (new_loc != NULL);
+	if (!new_loc)
+		return NULL;
+	new_uri = soup_uri_new_with_base (soup_message_get_uri (msg), new_loc);
+	if (!new_uri || !new_uri->host) {
+		if (new_uri)
+			soup_uri_free (new_uri);
+		return NULL;
+	}
 
+	return new_uri;
+}
+
+/**
+ * soup_session_would_redirect:
+ * @session: a #SoupSession
+ * @msg: a #SoupMessage that has response headers
+ *
+ * Checks if @msg contains a response that would cause @session to
+ * redirect it to a new URL (ignoring @msg's %SOUP_MESSAGE_NO_REDIRECT
+ * flag, and the number of times it has already been redirected).
+ *
+ * Return value: whether @msg would be redirected
+ *
+ * Since: 2.38
+ */
+gboolean
+soup_session_would_redirect (SoupSession *session, SoupMessage *msg)
+{
+	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
+	SoupURI *new_uri;
+
+	/* It must have an appropriate status code and method */
+	if (!SOUP_SESSION_WOULD_REDIRECT_AS_GET (session, msg) &&
+	    !SOUP_SESSION_WOULD_REDIRECT_AS_SAFE (session, msg))
+		return FALSE;
+
+	/* and a Location header that parses to an http URI */
+	if (!soup_message_headers_get_one (msg->response_headers, "Location"))
+		return FALSE;
+	new_uri = redirection_uri (msg);
+	if (!new_uri)
+		return FALSE;
+	if (!new_uri->host || !*new_uri->host ||
+	    (!uri_is_http (priv, new_uri) && !uri_is_https (priv, new_uri))) {
+		soup_uri_free (new_uri);
+		return FALSE;
+	}
+
+	soup_uri_free (new_uri);
+	return TRUE;
+}
+
+/**
+ * soup_session_redirect_message:
+ * @session: the session
+ * @msg: a #SoupMessage that has received a 3xx response
+ *
+ * Updates @msg's URI according to its status code and "Location"
+ * header, and requeues it on @session. Use this when you have set
+ * %SOUP_MESSAGE_NO_REDIRECT on a message, but have decided to allow a
+ * particular redirection to occur, or if you want to allow a
+ * redirection that #SoupSession will not perform automatically (eg,
+ * redirecting a non-safe method such as DELETE).
+ *
+ * If @msg's status code indicates that it should be retried as a GET
+ * request, then @msg will be modified accordingly.
+ *
+ * If @msg has already been redirected too many times, this will
+ * cause it to fail with %SOUP_STATUS_TOO_MANY_REDIRECTS.
+ *
+ * Return value: %TRUE if a redirection was applied, %FALSE if not
+ * (eg, because there was no Location header, or it could not be
+ * parsed).
+ *
+ * Since: 2.38
+ */
+gboolean
+soup_session_redirect_message (SoupSession *session, SoupMessage *msg)
+{
+	SoupMessageQueueItem *item;
+	SoupURI *new_uri;
+
+	new_uri = redirection_uri (msg);
+	if (!new_uri)
+		return FALSE;
+
+	item = soup_message_queue_lookup (soup_session_get_queue (session), msg);
+	if (!item)
+		return FALSE;
 	if (item->redirection_count >= SOUP_SESSION_MAX_REDIRECTION_COUNT) {
 		soup_session_cancel_message (session, msg, SOUP_STATUS_TOO_MANY_REDIRECTS);
-		return;
+		soup_message_queue_item_unref (item);
+		return FALSE;
 	}
 	item->redirection_count++;
+	soup_message_queue_item_unref (item);
 
-	if (msg->status_code == SOUP_STATUS_SEE_OTHER ||
-	    (msg->status_code == SOUP_STATUS_FOUND &&
-	     !SOUP_METHOD_IS_SAFE (msg->method)) ||
-	    (msg->status_code == SOUP_STATUS_MOVED_PERMANENTLY &&
-	     msg->method == SOUP_METHOD_POST)) {
+	if (SOUP_SESSION_WOULD_REDIRECT_AS_GET (session, msg)) {
 		if (msg->method != SOUP_METHOD_HEAD) {
-			/* Redirect using a GET */
 			g_object_set (msg,
 				      SOUP_MESSAGE_METHOD, SOUP_METHOD_GET,
 				      NULL);
@@ -1280,50 +1568,39 @@ redirect_handler (SoupMessage *msg, gpointer user_data)
 					  SOUP_MEMORY_STATIC, NULL, 0);
 		soup_message_headers_set_encoding (msg->request_headers,
 						   SOUP_ENCODING_NONE);
-	} else if (msg->status_code == SOUP_STATUS_MOVED_PERMANENTLY ||
-		   msg->status_code == SOUP_STATUS_TEMPORARY_REDIRECT ||
-		   msg->status_code == SOUP_STATUS_FOUND) {
-		/* Don't redirect non-safe methods */
-		if (!SOUP_METHOD_IS_SAFE (msg->method))
-			return;
-	} else {
-		/* Three possibilities:
-		 *
-		 *   1) This was a non-3xx response that happened to
-		 *      have a "Location" header
-		 *   2) It's a non-redirecty 3xx response (300, 304,
-		 *      305, 306)
-		 *   3) It's some newly-defined 3xx response (308+)
-		 *
-		 * We ignore all of these cases. In the first two,
-		 * redirecting would be explicitly wrong, and in the
-		 * last case, we have no clue if the 3xx response is
-		 * supposed to be redirecty or non-redirecty. Plus,
-		 * 2616 says unrecognized status codes should be
-		 * treated as the equivalent to the x00 code, and we
-		 * don't redirect on 300, so therefore we shouldn't
-		 * redirect on 308+ either.
-		 */
-		return;
-	}
-
-	/* Location is supposed to be an absolute URI, but some sites
-	 * are lame, so we use soup_uri_new_with_base().
-	 */
-	new_uri = soup_uri_new_with_base (soup_message_get_uri (msg), new_loc);
-	if (!new_uri || !new_uri->host) {
-		if (new_uri)
-			soup_uri_free (new_uri);
-		soup_message_set_status_full (msg,
-					      SOUP_STATUS_MALFORMED,
-					      "Invalid Redirect URL");
-		return;
 	}
 
 	soup_message_set_uri (msg, new_uri);
 	soup_uri_free (new_uri);
 
 	soup_session_requeue_message (session, msg);
+	return TRUE;
+}
+
+static void
+redirect_handler (SoupMessage *msg, gpointer user_data)
+{
+	SoupMessageQueueItem *item = user_data;
+	SoupSession *session = item->session;
+
+	if (!soup_session_would_redirect (session, msg)) {
+		SoupURI *new_uri = redirection_uri (msg);
+		gboolean invalid = !new_uri || !new_uri->host;
+
+		if (new_uri)
+			soup_uri_free (new_uri);
+		if (invalid) {
+			/* Really we should just leave the status as-is,
+			 * but that would be an API break.
+			 */
+			soup_message_set_status_full (msg,
+						      SOUP_STATUS_MALFORMED,
+						      "Invalid Redirect URL");
+		}
+		return;
+	}
+
+	soup_session_redirect_message (session, msg);
 }
 
 void
@@ -1544,17 +1821,18 @@ soup_session_get_connection (SoupSession *session,
 	}
 
 	uri = soup_message_get_uri (item->msg);
-	if (uri->scheme == SOUP_URI_SCHEME_HTTPS && item->proxy_addr)
+	if (uri_is_https (priv, uri) && item->proxy_addr)
 		tunnel_addr = host->addr;
 
 	conn = soup_connection_new (
 		SOUP_CONNECTION_REMOTE_ADDRESS, remote_addr,
 		SOUP_CONNECTION_TUNNEL_ADDRESS, tunnel_addr,
 		SOUP_CONNECTION_PROXY_URI, item->proxy_uri,
-		SOUP_CONNECTION_SSL, uri->scheme == SOUP_URI_SCHEME_HTTPS,
+		SOUP_CONNECTION_SSL, uri_is_https (priv, uri),
 		SOUP_CONNECTION_SSL_CREDENTIALS, priv->tlsdb,
 		SOUP_CONNECTION_SSL_STRICT, (priv->tlsdb != NULL) && priv->ssl_strict,
 		SOUP_CONNECTION_ASYNC_CONTEXT, priv->async_context,
+		SOUP_CONNECTION_USE_THREAD_CONTEXT, priv->use_thread_context,
 		SOUP_CONNECTION_TIMEOUT, priv->io_timeout,
 		SOUP_CONNECTION_IDLE_TIMEOUT, priv->idle_timeout,
 		SOUP_CONNECTION_SSL_FALLBACK, host->ssl_fallback,
@@ -2005,7 +2283,8 @@ soup_session_prepare_for_uri (SoupSession *session, SoupURI *uri)
 	addr = g_object_ref (host->addr);
 	g_mutex_unlock (priv->host_lock);
 
-	soup_address_resolve_async (addr, priv->async_context,
+	soup_address_resolve_async (addr,
+				    soup_session_get_async_context (session),
 				    NULL, NULL, NULL);
 	g_object_unref (addr);
 }
