@@ -32,6 +32,10 @@ typedef struct {
 	 */
 	const char *provided;
 
+	/* Whether to pass user and password in the URL or not.
+	 */
+	gboolean url_auth;
+
 	/* Expected passwords, 1 character each. (As with the provided
 	 * passwords, with the addition that '0' means "no
 	 * Authorization header expected".) Used to verify that soup
@@ -43,116 +47,119 @@ typedef struct {
 	guint final_status;
 } SoupAuthTest;
 
-static SoupAuthTest tests[] = {
+/* Will either point to main_tests or relogin_tests
+ */
+static SoupAuthTest *current_tests;
+
+static SoupAuthTest main_tests[] = {
 	{ "No auth available, should fail",
-	  "Basic/realm1/", "", "0", SOUP_STATUS_UNAUTHORIZED },
+	  "Basic/realm1/", "", FALSE, "0", SOUP_STATUS_UNAUTHORIZED },
 
 	{ "Should fail with no auth, fail again with bad password, and give up",
-	  "Basic/realm2/", "4", "04", SOUP_STATUS_UNAUTHORIZED },
+	  "Basic/realm2/", "4", FALSE, "04", SOUP_STATUS_UNAUTHORIZED },
 
 	{ "Auth provided this time, so should succeed",
-	  "Basic/realm1/", "1", "01", SOUP_STATUS_OK },
+	  "Basic/realm1/", "1", FALSE, "01", SOUP_STATUS_OK },
 
 	{ "Now should automatically reuse previous auth",
-	  "Basic/realm1/", "", "1", SOUP_STATUS_OK },
+	  "Basic/realm1/", "", FALSE, "1", SOUP_STATUS_OK },
 
 	{ "Subdir should also automatically reuse auth",
-	  "Basic/realm1/subdir/", "", "1", SOUP_STATUS_OK },
+	  "Basic/realm1/subdir/", "", FALSE, "1", SOUP_STATUS_OK },
 
 	{ "Subdir should retry last auth, but will fail this time",
-	  "Basic/realm1/realm2/", "", "1", SOUP_STATUS_UNAUTHORIZED },
+	  "Basic/realm1/realm2/", "", FALSE, "1", SOUP_STATUS_UNAUTHORIZED },
 
 	{ "Now should use provided auth",
-	  "Basic/realm1/realm2/", "2", "02", SOUP_STATUS_OK },
+	  "Basic/realm1/realm2/", "2", FALSE, "02", SOUP_STATUS_OK },
 
 	{ "Reusing last auth. Should succeed on first try",
-	  "Basic/realm1/realm2/", "", "2", SOUP_STATUS_OK },
+	  "Basic/realm1/realm2/", "", FALSE, "2", SOUP_STATUS_OK },
 
 	{ "Reuse will fail, but 2nd try will succeed because it's a known realm",
-	  "Basic/realm1/realm2/realm1/", "", "21", SOUP_STATUS_OK },
+	  "Basic/realm1/realm2/realm1/", "", FALSE, "21", SOUP_STATUS_OK },
 
 	{ "Should succeed on first try. (Known realm with cached password)",
-	  "Basic/realm2/", "", "2", SOUP_STATUS_OK },
+	  "Basic/realm2/", "", FALSE, "2", SOUP_STATUS_OK },
 
 	{ "Fail once, then use typoed password, then use right password",
-	  "Basic/realm3/", "43", "043", SOUP_STATUS_OK },
+	  "Basic/realm3/", "43", FALSE, "043", SOUP_STATUS_OK },
 
 
 	{ "No auth available, should fail",
-	  "Digest/realm1/", "", "0", SOUP_STATUS_UNAUTHORIZED },
+	  "Digest/realm1/", "", FALSE, "0", SOUP_STATUS_UNAUTHORIZED },
 
 	{ "Should fail with no auth, fail again with bad password, and give up",
-	  "Digest/realm2/", "4", "04", SOUP_STATUS_UNAUTHORIZED },
+	  "Digest/realm2/", "4", FALSE, "04", SOUP_STATUS_UNAUTHORIZED },
 
 	{ "Known realm, auth provided, so should succeed",
-	  "Digest/realm1/", "1", "01", SOUP_STATUS_OK },
+	  "Digest/realm1/", "1", FALSE, "01", SOUP_STATUS_OK },
 
 	{ "Now should automatically reuse previous auth",
-	  "Digest/realm1/", "", "1", SOUP_STATUS_OK },
+	  "Digest/realm1/", "", FALSE, "1", SOUP_STATUS_OK },
 
 	{ "Subdir should also automatically reuse auth",
-	  "Digest/realm1/subdir/", "", "1", SOUP_STATUS_OK },
+	  "Digest/realm1/subdir/", "", FALSE, "1", SOUP_STATUS_OK },
 
 	{ "Password provided, should succeed",
-	  "Digest/realm2/", "2", "02", SOUP_STATUS_OK },
+	  "Digest/realm2/", "2", FALSE, "02", SOUP_STATUS_OK },
 
 	{ "Should already know correct domain and use provided auth on first try",
-	  "Digest/realm1/realm2/", "2", "2", SOUP_STATUS_OK },
+	  "Digest/realm1/realm2/", "2", FALSE, "2", SOUP_STATUS_OK },
 
 	{ "Reusing last auth. Should succeed on first try",
-	  "Digest/realm1/realm2/", "", "2", SOUP_STATUS_OK },
+	  "Digest/realm1/realm2/", "", FALSE, "2", SOUP_STATUS_OK },
 
 	{ "Should succeed on first try because of earlier domain directive",
-	  "Digest/realm1/realm2/realm1/", "", "1", SOUP_STATUS_OK },
+	  "Digest/realm1/realm2/realm1/", "", FALSE, "1", SOUP_STATUS_OK },
 
 	{ "Fail once, then use typoed password, then use right password",
-	  "Digest/realm3/", "43", "043", SOUP_STATUS_OK },
+	  "Digest/realm3/", "43", FALSE, "043", SOUP_STATUS_OK },
 
 
 	{ "Make sure we haven't forgotten anything",
-	  "Basic/realm1/", "", "1", SOUP_STATUS_OK },
+	  "Basic/realm1/", "", FALSE, "1", SOUP_STATUS_OK },
 
 	{ "Make sure we haven't forgotten anything",
-	  "Basic/realm1/realm2/", "", "2", SOUP_STATUS_OK },
+	  "Basic/realm1/realm2/", "", FALSE, "2", SOUP_STATUS_OK },
 
 	{ "Make sure we haven't forgotten anything",
-	  "Basic/realm1/realm2/realm1/", "", "1", SOUP_STATUS_OK },
+	  "Basic/realm1/realm2/realm1/", "", FALSE, "1", SOUP_STATUS_OK },
 
 	{ "Make sure we haven't forgotten anything",
-	  "Basic/realm2/", "", "2", SOUP_STATUS_OK },
+	  "Basic/realm2/", "", FALSE, "2", SOUP_STATUS_OK },
 
 	{ "Make sure we haven't forgotten anything",
-	  "Basic/realm3/", "", "3", SOUP_STATUS_OK },
+	  "Basic/realm3/", "", FALSE, "3", SOUP_STATUS_OK },
 
 
 	{ "Make sure we haven't forgotten anything",
-	  "Digest/realm1/", "", "1", SOUP_STATUS_OK },
+	  "Digest/realm1/", "", FALSE, "1", SOUP_STATUS_OK },
 
 	{ "Make sure we haven't forgotten anything",
-	  "Digest/realm1/realm2/", "", "2", SOUP_STATUS_OK },
+	  "Digest/realm1/realm2/", "", FALSE, "2", SOUP_STATUS_OK },
 
 	{ "Make sure we haven't forgotten anything",
-	  "Digest/realm1/realm2/realm1/", "", "1", SOUP_STATUS_OK },
+	  "Digest/realm1/realm2/realm1/", "", FALSE, "1", SOUP_STATUS_OK },
 
 	{ "Make sure we haven't forgotten anything",
-	  "Digest/realm2/", "", "2", SOUP_STATUS_OK },
+	  "Digest/realm2/", "", FALSE, "2", SOUP_STATUS_OK },
 
 	{ "Make sure we haven't forgotten anything",
-	  "Digest/realm3/", "", "3", SOUP_STATUS_OK },
+	  "Digest/realm3/", "", FALSE, "3", SOUP_STATUS_OK },
 
 	{ "Now the server will reject the formerly-good password",
-	  "Basic/realm1/not/", "1" /* should not be used */, "1", SOUP_STATUS_UNAUTHORIZED },
+	  "Basic/realm1/not/", "1", FALSE, /* should not be used */ "1", SOUP_STATUS_UNAUTHORIZED },
 
 	{ "Make sure we've forgotten it",
-	  "Basic/realm1/", "", "0", SOUP_STATUS_UNAUTHORIZED },
+	  "Basic/realm1/", "", FALSE, "0", SOUP_STATUS_UNAUTHORIZED },
 
 	{ "Likewise, reject the formerly-good Digest password",
-	  "Digest/realm1/not/", "1" /* should not be used */, "1", SOUP_STATUS_UNAUTHORIZED },
+	  "Digest/realm1/not/", "1", FALSE, /* should not be used */ "1", SOUP_STATUS_UNAUTHORIZED },
 
 	{ "Make sure we've forgotten it",
-	  "Digest/realm1/", "", "0", SOUP_STATUS_UNAUTHORIZED }
+	  "Digest/realm1/", "", FALSE, "0", SOUP_STATUS_UNAUTHORIZED }
 };
-static int ntests = sizeof (tests) / sizeof (tests[0]);
 
 static const char *auths[] = {
 	"no password", "password 1",
@@ -226,14 +233,14 @@ authenticate (SoupSession *session, SoupMessage *msg,
 	char *username, *password;
 	char num;
 
-	if (!tests[*i].provided[0])
+	if (!current_tests[*i].provided[0])
 		return;
 	if (retrying) {
-		if (!tests[*i].provided[1])
+		if (!current_tests[*i].provided[1])
 			return;
-		num = tests[*i].provided[1];
+		num = current_tests[*i].provided[1];
 	} else
-		num = tests[*i].provided[0];
+		num = current_tests[*i].provided[0];
 
 	username = g_strdup_printf ("user%c", num);
 	password = g_strdup_printf ("realm%c", num);
@@ -296,6 +303,111 @@ bug271540_finished (SoupSession *session, SoupMessage *msg, gpointer data)
 		g_main_loop_quit (loop);
 }
 
+static void
+do_pipelined_auth_test (const char *base_uri)
+{
+	SoupSession *session;
+	SoupMessage *msg;
+	gboolean authenticated;
+	char *uri;
+	int i;
+
+	debug_printf (1, "Testing pipelined auth (bug 271540):\n");
+	session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC, NULL);
+
+	authenticated = FALSE;
+	g_signal_connect (session, "authenticate",
+			  G_CALLBACK (bug271540_authenticate), &authenticated);
+
+	uri = g_strconcat (base_uri, "Basic/realm1/", NULL);
+	for (i = 0; i < 10; i++) {
+		msg = soup_message_new (SOUP_METHOD_GET, uri);
+		g_object_set_data (G_OBJECT (msg), "#", GINT_TO_POINTER (i + 1));
+		g_signal_connect (msg, "wrote_headers",
+				  G_CALLBACK (bug271540_sent), &authenticated);
+
+		soup_session_queue_message (session, msg,
+					    bug271540_finished, &i);
+	}
+	g_free (uri);
+
+	loop = g_main_loop_new (NULL, TRUE);
+	g_main_loop_run (loop);
+	g_main_loop_unref (loop);
+
+	soup_test_session_abort_unref (session);
+}
+
+/* We test two different things here:
+ *
+ *   1. If we get a 401 response with "WWW-Authenticate: Digest
+ *      stale=true...", we should retry and succeed *without* the
+ *      session asking for a password again.
+ *
+ *   2. If we get a successful response with "Authentication-Info:
+ *      nextnonce=...", we should update the nonce automatically so as
+ *      to avoid getting a stale nonce error on the next request.
+ *
+ * In our Apache config, /Digest/realm1 and /Digest/realm1/expire are
+ * set up to use the same auth info, but only the latter has an
+ * AuthDigestNonceLifetime (of 2 seconds). The way nonces work in
+ * Apache, a nonce received from /Digest/realm1 will still expire in
+ * /Digest/realm1/expire, but it won't issue a nextnonce for a request
+ * in /Digest/realm1. This lets us test both behaviors.
+ *
+ * The expected conversation is:
+ *
+ * First message
+ *   GET /Digest/realm1
+ *
+ *   401 Unauthorized
+ *   WWW-Authenticate: Digest nonce=A
+ *
+ *   [emit 'authenticate']
+ *
+ *   GET /Digest/realm1
+ *   Authorization: Digest nonce=A
+ *
+ *   200 OK
+ *   [No Authentication-Info]
+ *
+ * [sleep 2 seconds: nonce A is no longer valid, but we have no
+ * way of knowing that]
+ *
+ * Second message
+ *   GET /Digest/realm1/expire/
+ *   Authorization: Digest nonce=A
+ *
+ *   401 Unauthorized
+ *   WWW-Authenticate: Digest stale=true nonce=B
+ *
+ *   GET /Digest/realm1/expire/
+ *   Authorization: Digest nonce=B
+ *
+ *   200 OK
+ *   Authentication-Info: nextnonce=C
+ *
+ * [sleep 1 second]
+ *
+ * Third message
+ *   GET /Digest/realm1/expire/
+ *   Authorization: Digest nonce=C
+ *   [nonce=B would work here too]
+ *
+ *   200 OK
+ *   Authentication-Info: nextnonce=D
+ *
+ * [sleep 1 second; nonces B and C are no longer valid, but D is]
+ *
+ * Fourth message
+ *   GET /Digest/realm1/expire/
+ *   Authorization: Digest nonce=D
+ *
+ *   200 OK
+ *   Authentication-Info: nextnonce=D
+ *
+ */
+
 static void
 digest_nonce_authenticate (SoupSession *session, SoupMessage *msg,
 			   SoupAuth *auth, gboolean retrying, gpointer data)
@@ -352,6 +464,31 @@ do_digest_nonce_test (SoupSession *session,
 	g_object_unref (msg);
 }
 
+static void
+do_digest_expiration_test (const char *base_uri)
+{
+	SoupSession *session;
+	char *uri;
+
+	debug_printf (1, "\nTesting digest nonce expiration:\n");
+
+	session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC, NULL);
+
+	uri = g_strconcat (base_uri, "Digest/realm1/", NULL);
+	do_digest_nonce_test (session, "First", uri, TRUE, TRUE);
+	g_free (uri);
+	sleep (2);
+	uri = g_strconcat (base_uri, "Digest/realm1/expire/", NULL);
+	do_digest_nonce_test (session, "Second", uri, TRUE, FALSE);
+	sleep (1);
+	do_digest_nonce_test (session, "Third", uri, FALSE, FALSE);
+	sleep (1);
+	do_digest_nonce_test (session, "Fourth", uri, FALSE, FALSE);
+	g_free (uri);
+
+	soup_test_session_abort_unref (session);
+}
+
 /* Async auth test. We queue three requests to /Basic/realm1, ensuring
  * that they are sent in order. The first and third ones will be
  * paused from the authentication callback. The second will be allowed
@@ -443,6 +580,7 @@ do_async_auth_test (const char *base_uri)
 
 	debug_printf (1, "\nTesting async auth:\n");
 
+	loop = g_main_loop_new (NULL, TRUE);
 	session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC, NULL);
 	remaining = 0;
 
@@ -608,6 +746,7 @@ do_async_auth_test (const char *base_uri)
 	g_object_unref (msg1);
 
 	g_free (uri);
+	g_main_loop_unref (loop);
 }
 
 typedef struct {
@@ -646,7 +785,8 @@ select_auth_authenticate (SoupSession *session, SoupMessage *msg,
 }
 
 static void
-select_auth_test_one (SoupURI *uri, const char *password,
+select_auth_test_one (SoupURI *uri,
+		      gboolean disable_digest, const char *password,
 		      const char *first_headers, const char *first_response,
 		      const char *second_headers, const char *second_response,
 		      guint final_status)
@@ -656,6 +796,9 @@ select_auth_test_one (SoupURI *uri, const char *password,
 	SoupSession *session;
 
 	session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC, NULL);
+	if (disable_digest)
+		soup_session_remove_feature_by_type (session, SOUP_TYPE_AUTH_DIGEST);
+
 	g_signal_connect (session, "authenticate",
 			  G_CALLBACK (select_auth_authenticate), &sad);
 	memset (&sad, 0, sizeof (sad));
@@ -719,7 +862,9 @@ static gboolean
 server_basic_auth_callback (SoupAuthDomain *auth_domain, SoupMessage *msg,
 			    const char *username, const char *password, gpointer data)
 {
-	return FALSE;
+	if (strcmp (username, "user") != 0)
+		return FALSE;
+	return strcmp (password, "good-basic") == 0;
 }
 
 static char *
@@ -769,28 +914,43 @@ do_select_auth_test (void)
 		NULL);
 	soup_server_add_auth_domain (server, digest_auth_domain);
 
-	/* FIXME: when we support disabling auth types in the session,
-	 * test that too.
-	 */
-
 	debug_printf (1, "  Testing with no auth\n");
-	select_auth_test_one (uri, NULL,
+	select_auth_test_one (uri, FALSE, NULL,
 			      "Basic, Digest", "Digest",
 			      NULL, NULL,
 			      SOUP_STATUS_UNAUTHORIZED);
 
 	debug_printf (1, "  Testing with bad password\n");
-	select_auth_test_one (uri, "bad",
+	select_auth_test_one (uri, FALSE, "bad",
 			      "Basic, Digest", "Digest",
 			      "Basic, Digest", "Digest",
 			      SOUP_STATUS_UNAUTHORIZED);
 
 	debug_printf (1, "  Testing with good password\n");
-	select_auth_test_one (uri, "good",
+	select_auth_test_one (uri, FALSE, "good",
 			      "Basic, Digest", "Digest",
 			      NULL, NULL,
 			      SOUP_STATUS_OK);
 
+	/* Test with Digest disabled in the client. */
+	debug_printf (1, "  Testing without Digest with no auth\n");
+	select_auth_test_one (uri, TRUE, NULL,
+			      "Basic, Digest", "Basic",
+			      NULL, NULL,
+			      SOUP_STATUS_UNAUTHORIZED);
+
+	debug_printf (1, "  Testing without Digest with bad password\n");
+	select_auth_test_one (uri, TRUE, "bad",
+			      "Basic, Digest", "Basic",
+			      "Basic, Digest", "Basic",
+			      SOUP_STATUS_UNAUTHORIZED);
+
+	debug_printf (1, "  Testing without Digest with good password\n");
+	select_auth_test_one (uri, TRUE, "good-basic",
+			      "Basic, Digest", "Basic",
+			      NULL, NULL,
+			      SOUP_STATUS_OK);
+
 	/* Now flip the order of the domains, verify that this flips
 	 * the order of the headers, and make sure that digest auth
 	 * *still* gets used.
@@ -802,19 +962,19 @@ do_select_auth_test (void)
 	soup_server_add_auth_domain (server, basic_auth_domain);
 
 	debug_printf (1, "  Testing flipped with no auth\n");
-	select_auth_test_one (uri, NULL,
+	select_auth_test_one (uri, FALSE, NULL,
 			      "Digest, Basic", "Digest",
 			      NULL, NULL,
 			      SOUP_STATUS_UNAUTHORIZED);
 
 	debug_printf (1, "  Testing flipped with bad password\n");
-	select_auth_test_one (uri, "bad",
+	select_auth_test_one (uri, FALSE, "bad",
 			      "Digest, Basic", "Digest",
 			      "Digest, Basic", "Digest",
 			      SOUP_STATUS_UNAUTHORIZED);
 
 	debug_printf (1, "  Testing flipped with good password\n");
-	select_auth_test_one (uri, "good",
+	select_auth_test_one (uri, FALSE, "good",
 			      "Digest, Basic", "Digest",
 			      NULL, NULL,
 			      SOUP_STATUS_OK);
@@ -825,39 +985,164 @@ do_select_auth_test (void)
 	soup_test_server_quit_unref (server);
 }
 
-int
-main (int argc, char **argv)
+static void
+sneakily_close_connection (SoupMessage *msg, gpointer user_data)
+{
+	/* Sneakily close the connection after the response, by
+	 * tricking soup-message-io into thinking that had been
+	 * the plan all along.
+	 */
+	soup_message_headers_append (msg->response_headers,
+				     "Connection", "close");
+}
+
+static void
+auth_close_request_started (SoupServer *server, SoupMessage *msg,
+			    SoupClientContext *client, gpointer user_data)
+{
+	g_signal_connect (msg, "wrote-headers",
+			  G_CALLBACK (sneakily_close_connection), NULL);
+}
+
+typedef struct {
+	SoupSession *session;
+	SoupMessage *msg;
+	SoupAuth *auth;
+} AuthCloseData;
+
+static gboolean
+auth_close_idle_authenticate (gpointer user_data)
+{
+	AuthCloseData *acd = user_data;
+
+	soup_auth_authenticate (acd->auth, "user", "good-basic");
+	soup_session_unpause_message (acd->session, acd->msg);
+
+	g_object_unref (acd->auth);
+	return FALSE;
+}
+
+static void
+auth_close_authenticate (SoupSession *session, SoupMessage *msg,
+			 SoupAuth *auth, gboolean retrying, gpointer data)
+{
+	AuthCloseData *acd = data;
+
+	soup_session_pause_message (session, msg);
+	acd->auth = g_object_ref (auth);
+	g_idle_add (auth_close_idle_authenticate, acd);
+}
+
+static void
+do_auth_close_test (void)
+{
+	SoupServer *server;
+	SoupAuthDomain *basic_auth_domain;
+	SoupURI *uri;
+	AuthCloseData acd;
+
+	debug_printf (1, "\nTesting auth when server times out connection:\n");
+
+	server = soup_test_server_new (FALSE);
+	soup_server_add_handler (server, NULL,
+				 server_callback, NULL, NULL);
+
+	uri = soup_uri_new ("http://127.0.0.1/close");
+	soup_uri_set_port (uri, soup_server_get_port (server));
+
+	basic_auth_domain = soup_auth_domain_basic_new (
+		SOUP_AUTH_DOMAIN_REALM, "auth-test",
+		SOUP_AUTH_DOMAIN_ADD_PATH, "/",
+		SOUP_AUTH_DOMAIN_BASIC_AUTH_CALLBACK, server_basic_auth_callback,
+		NULL);
+	soup_server_add_auth_domain (server, basic_auth_domain);
+
+	g_signal_connect (server, "request-started",
+			  G_CALLBACK (auth_close_request_started), NULL);
+
+	acd.session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC, NULL);
+	g_signal_connect (acd.session, "authenticate",
+			  G_CALLBACK (auth_close_authenticate), &acd);
+
+	acd.msg = soup_message_new_from_uri ("GET", uri);
+	soup_session_send_message (acd.session, acd.msg);
+
+	if (acd.msg->status_code != SOUP_STATUS_OK) {
+		debug_printf (1, "    Final status wrong: expected %u, got %u %s\n",
+			      SOUP_STATUS_OK, acd.msg->status_code,
+			      acd.msg->reason_phrase);
+		errors++;
+	}
+
+	g_object_unref (acd.msg);
+	soup_test_session_abort_unref (acd.session);
+	soup_test_server_quit_unref (server);
+}
+
+static SoupAuthTest relogin_tests[] = {
+	{ "Auth provided via URL, should succeed",
+	  "Basic/realm12/", "1", TRUE, "01", SOUP_STATUS_OK },
+
+	{ "Now should automatically reuse previous auth",
+	  "Basic/realm12/", "", FALSE, "1", SOUP_STATUS_OK },
+
+	{ "Different auth provided via URL for the same realm, should succeed",
+	  "Basic/realm12/", "2", TRUE, "2", SOUP_STATUS_OK },
+
+	{ "Subdir should also automatically reuse auth",
+	  "Basic/realm12/subdir/", "", FALSE, "2", SOUP_STATUS_OK },
+
+	{ "Should fail with no auth",
+	  "Basic/realm12/", "4", TRUE, "4", SOUP_STATUS_UNAUTHORIZED },
+
+	{ "Make sure we've forgotten it",
+	  "Basic/realm12/", "", FALSE, "0", SOUP_STATUS_UNAUTHORIZED },
+
+	{ "Should fail with no auth, fail again with bad password, and give up",
+	  "Basic/realm12/", "3", FALSE, "03", SOUP_STATUS_UNAUTHORIZED },
+};
+
+static void
+do_batch_tests (const gchar *base_uri_str, gint ntests)
 {
 	SoupSession *session;
 	SoupMessage *msg;
-	const char *base_uri;
-	char *uri, *expected;
-	gboolean authenticated;
+	char *expected, *uristr;
+	SoupURI *base_uri;
 	int i;
 
-	test_init (argc, argv, NULL);
-	apache_init ();
-
-	base_uri = "http://127.0.0.1:47524/";
-
 	session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC, NULL);
 	g_signal_connect (session, "authenticate",
 			  G_CALLBACK (authenticate), &i);
 
+	base_uri = soup_uri_new (base_uri_str);
+
 	for (i = 0; i < ntests; i++) {
-		debug_printf (1, "Test %d: %s\n", i + 1, tests[i].explanation);
+		SoupURI *soup_uri = soup_uri_new_with_base (base_uri, current_tests[i].url);
 
-		uri = g_strconcat (base_uri, tests[i].url, NULL);
-		debug_printf (1, "  GET %s\n", uri);
+		debug_printf (1, "Test %d: %s\n", i + 1, current_tests[i].explanation);
 
-		msg = soup_message_new (SOUP_METHOD_GET, uri);
-		g_free (uri);
+		if (current_tests[i].url_auth) {
+			gchar *username = g_strdup_printf ("user%c", current_tests[i].provided[0]);
+			gchar *password = g_strdup_printf ("realm%c", current_tests[i].provided[0]);
+			soup_uri_set_user (soup_uri, username);
+			soup_uri_set_password (soup_uri, password);
+			g_free (username);
+			g_free (password);
+		}
+
+		msg = soup_message_new_from_uri (SOUP_METHOD_GET, soup_uri);
+		soup_uri_free (soup_uri);
 		if (!msg) {
 			fprintf (stderr, "auth-test: Could not parse URI\n");
 			exit (1);
 		}
 
-		expected = g_strdup (tests[i].expected);
+		uristr = soup_uri_to_string (soup_message_get_uri (msg), FALSE);
+		debug_printf (1, "  GET %s\n", uristr);
+		g_free (uristr);
+
+		expected = g_strdup (current_tests[i].expected);
 		soup_message_add_status_code_handler (
 			msg, "got_headers", SOUP_STATUS_UNAUTHORIZED,
 			G_CALLBACK (handler), expected);
@@ -878,141 +1163,47 @@ main (int argc, char **argv)
 		}
 		g_free (expected);
 
-		if (msg->status_code != tests[i].final_status) {
+		if (msg->status_code != current_tests[i].final_status) {
 			debug_printf (1, "  expected %d\n",
-				      tests[i].final_status);
+				      current_tests[i].final_status);
 		}
 
 		debug_printf (1, "\n");
 
 		g_object_unref (msg);
 	}
-	soup_test_session_abort_unref (session);
-
-	/* And now for some regression tests */
-	loop = g_main_loop_new (NULL, TRUE);
-
-	debug_printf (1, "Testing pipelined auth (bug 271540):\n");
-	session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC, NULL);
-
-	authenticated = FALSE;
-	g_signal_connect (session, "authenticate",
-			  G_CALLBACK (bug271540_authenticate), &authenticated);
-
-	uri = g_strconcat (base_uri, "Basic/realm1/", NULL);
-	for (i = 0; i < 10; i++) {
-		msg = soup_message_new (SOUP_METHOD_GET, uri);
-		g_object_set_data (G_OBJECT (msg), "#", GINT_TO_POINTER (i + 1));
-		g_signal_connect (msg, "wrote_headers",
-				  G_CALLBACK (bug271540_sent), &authenticated);
-
-		soup_session_queue_message (session, msg,
-					    bug271540_finished, &i);
-	}
-	g_free (uri);
-
-	g_main_loop_run (loop);
-	soup_test_session_abort_unref (session);
-
-	debug_printf (1, "\nTesting digest nonce expiration:\n");
-
-	/* We test two different things here:
-	 *
-	 *   1. If we get a 401 response with
-	 *      "WWW-Authenticate: Digest stale=true...", we should
-	 *      retry and succeed *without* the session asking for a
-	 *      password again.
-	 *
-	 *   2. If we get a successful response with
-	 *      "Authentication-Info: nextnonce=...", we should update
-	 *      the nonce automatically so as to avoid getting a
-	 *      stale nonce error on the next request.
-	 *
-	 * In our Apache config, /Digest/realm1 and
-	 * /Digest/realm1/expire are set up to use the same auth info,
-	 * but only the latter has an AuthDigestNonceLifetime (of 2
-	 * seconds). The way nonces work in Apache, a nonce received
-	 * from /Digest/realm1 will still expire in
-	 * /Digest/realm1/expire, but it won't issue a nextnonce for a
-	 * request in /Digest/realm1. This lets us test both
-	 * behaviors.
-	 *
-	 * The expected conversation is:
-	 *
-	 * First message
-	 *   GET /Digest/realm1
-	 *
-	 *   401 Unauthorized
-	 *   WWW-Authenticate: Digest nonce=A
-	 *
-	 *   [emit 'authenticate']
-	 *
-	 *   GET /Digest/realm1
-	 *   Authorization: Digest nonce=A
-	 *
-	 *   200 OK
-	 *   [No Authentication-Info]
-	 *
-	 * [sleep 2 seconds: nonce A is no longer valid, but we have no
-	 * way of knowing that]
-	 *
-	 * Second message
-	 *   GET /Digest/realm1/expire/
-	 *   Authorization: Digest nonce=A
-	 *
-	 *   401 Unauthorized
-	 *   WWW-Authenticate: Digest stale=true nonce=B
-	 *
-	 *   GET /Digest/realm1/expire/
-	 *   Authorization: Digest nonce=B
-	 *
-	 *   200 OK
-	 *   Authentication-Info: nextnonce=C
-	 *
-	 * [sleep 1 second]
-	 *
-	 * Third message
-	 *   GET /Digest/realm1/expire/
-	 *   Authorization: Digest nonce=C
-	 *   [nonce=B would work here too]
-	 *
-	 *   200 OK
-	 *   Authentication-Info: nextnonce=D
-	 *
-	 * [sleep 1 second; nonces B and C are no longer valid, but D is]
-	 *
-	 * Fourth message
-	 *   GET /Digest/realm1/expire/
-	 *   Authorization: Digest nonce=D
-	 *
-	 *   200 OK
-	 *   Authentication-Info: nextnonce=D
-	 *
-	 */
-
-	session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC, NULL);
-
-	uri = g_strconcat (base_uri, "Digest/realm1/", NULL);
-	do_digest_nonce_test (session, "First", uri, TRUE, TRUE);
-	g_free (uri);
-	sleep (2);
-	uri = g_strconcat (base_uri, "Digest/realm1/expire/", NULL);
-	do_digest_nonce_test (session, "Second", uri, TRUE, FALSE);
-	sleep (1);
-	do_digest_nonce_test (session, "Third", uri, FALSE, FALSE);
-	sleep (1);
-	do_digest_nonce_test (session, "Fourth", uri, FALSE, FALSE);
-	g_free (uri);
+	soup_uri_free (base_uri);
 
 	soup_test_session_abort_unref (session);
+}
 
-	/* Async auth */
+int
+main (int argc, char **argv)
+{
+	const char *base_uri;
+	int ntests;
+
+	test_init (argc, argv, NULL);
+	apache_init ();
+
+	base_uri = "http://127.0.0.1:47524/";
+
+	/* Main tests */
+	current_tests = main_tests;
+	ntests = G_N_ELEMENTS (main_tests);
+	do_batch_tests (base_uri, ntests);
+
+	/* Re-login tests */
+	current_tests = relogin_tests;
+	ntests = G_N_ELEMENTS (relogin_tests);
+	do_batch_tests (base_uri, ntests);
+
+	/* Other regression tests */
+	do_pipelined_auth_test (base_uri);
+	do_digest_expiration_test (base_uri);
 	do_async_auth_test (base_uri);
-
-	/* Selecting correct auth when multiple auth types are available */
 	do_select_auth_test ();
-
-	g_main_loop_unref (loop);
+	do_auth_close_test ();
 
 	test_cleanup ();
 	return errors != 0;
