@@ -20,6 +20,7 @@
 
 SoupServer *server;
 SoupURI *base_uri;
+GMutex *server_mutex;
 
 static gboolean
 auth_callback (SoupAuthDomain *auth_domain, SoupMessage *msg,
@@ -40,6 +41,12 @@ close_socket (SoupMessage *msg, gpointer user_data)
 	SoupSocket *sock = user_data;
 
 	soup_socket_disconnect (sock);
+
+	/* But also add the missing data to the message now, so
+	 * SoupServer can clean up after itself properly.
+	 */
+	soup_message_body_append (msg->response_body, SOUP_MEMORY_STATIC,
+				  "foo", 3);
 }
 
 static void
@@ -48,6 +55,62 @@ timeout_socket (SoupSocket *sock, gpointer user_data)
 	soup_socket_disconnect (sock);
 }
 
+static void
+timeout_request_started (SoupServer *server, SoupMessage *msg,
+			 SoupClientContext *client, gpointer user_data)
+{
+	SoupSocket *sock;
+	GMainContext *context = soup_server_get_async_context (server);
+	guint readable;
+
+	sock = soup_client_context_get_socket (client);
+	readable = g_signal_connect (sock, "readable",
+				    G_CALLBACK (timeout_socket), NULL);
+	while (soup_socket_is_connected (sock))
+		g_main_context_iteration (context, TRUE);
+	g_signal_handler_disconnect (sock, readable);
+	g_signal_handlers_disconnect_by_func (server, timeout_request_started, NULL);
+}
+
+static void
+setup_timeout_persistent (SoupServer *server, SoupSocket *sock)
+{
+	char buf[1];
+	gsize nread;
+
+	/* In order for the test to work correctly, we have to
+	 * close the connection *after* the client side writes
+	 * the request. To ensure that this happens reliably,
+	 * regardless of thread scheduling, we:
+	 *
+	 *   1. Try to read off the socket now, knowing it will
+	 *      fail (since the client is waiting for us to
+	 *      return a response). This will cause it to
+	 *      emit "readable" later.
+	 *   2. Connect to the server's request-started signal.
+	 *   3. Run an inner main loop from that signal handler
+	 *      until the socket emits "readable". (If we don't
+	 *      do this then it's possible the client's next
+	 *      request would be ready before we returned to
+	 *      the main loop, and so the signal would never be
+	 *      emitted.)
+	 *   4. Close the socket.
+	 */
+
+	soup_socket_read (sock, buf, 1, &nread, NULL, NULL);
+	g_signal_connect (server, "request-started",
+			  G_CALLBACK (timeout_request_started), NULL);
+}
+
+static gboolean
+timeout_finish_message (gpointer msg)
+{
+	SoupServer *server = g_object_get_data (G_OBJECT (msg), "server");
+
+	soup_server_unpause_message (server, msg);
+	return FALSE;
+}
+
 static void
 server_callback (SoupServer *server, SoupMessage *msg,
 		 const char *path, GHashTable *query,
@@ -55,6 +118,13 @@ server_callback (SoupServer *server, SoupMessage *msg,
 {
 	SoupURI *uri = soup_message_get_uri (msg);
 
+	/* The way this gets used in the tests, we don't actually
+	 * need to hold it through the whole function, so it's simpler
+	 * to just release it right away.
+	 */
+	g_mutex_lock (server_mutex);
+	g_mutex_unlock (server_mutex);
+
 	soup_message_headers_append (msg->response_headers,
 				     "X-Handled-By", "server_callback");
 
@@ -71,13 +141,7 @@ server_callback (SoupServer *server, SoupMessage *msg,
 	}
 
 	if (!strcmp (path, "/redirect")) {
-		soup_message_set_status (msg, SOUP_STATUS_FOUND);
-		soup_message_headers_append (msg->response_headers,
-					     /* Kids: don't try this at home!
-					      * RFC2616 says to use an
-					      * absolute URI!
-					      */
-					     "Location", "/");
+		soup_message_set_redirect (msg, SOUP_STATUS_FOUND, "/");
 		return;
 	}
 
@@ -119,12 +183,15 @@ server_callback (SoupServer *server, SoupMessage *msg,
 	if (!strcmp (path, "/timeout-persistent")) {
 		SoupSocket *sock;
 
-		/* We time out the persistent connection as soon as
-		 * the client starts writing the next request.
-		 */
 		sock = soup_client_context_get_socket (context);
-		g_signal_connect (sock, "readable",
-				  G_CALLBACK (timeout_socket), NULL);
+		setup_timeout_persistent (server, sock);
+	}
+
+	if (!strcmp (path, "/slow")) {
+		soup_server_pause_message (server, msg);
+		g_object_set_data (G_OBJECT (msg), "server", server);
+		soup_add_timeout (soup_server_get_async_context (server),
+				  1000, timeout_finish_message, msg);
 	}
 
 	soup_message_set_status (msg, SOUP_STATUS_OK);
@@ -751,6 +818,194 @@ do_persistent_connection_timeout_test (void)
 	soup_test_session_abort_unref (session);
 }
 
+static GMainLoop *max_conns_loop;
+static int msgs_done;
+#define MAX_CONNS 2
+#define TEST_CONNS (MAX_CONNS * 2)
+
+static gboolean
+idle_start_server (gpointer data)
+{
+	g_mutex_unlock (server_mutex);
+	return FALSE;
+}
+
+static gboolean
+quit_loop (gpointer data)
+{
+	g_main_loop_quit (max_conns_loop);
+	return FALSE;
+}
+
+static void
+max_conns_request_started (SoupSession *session, SoupMessage *msg,
+			   SoupSocket *socket, gpointer user_data)
+{
+	if (++msgs_done == MAX_CONNS)
+		g_timeout_add (100, quit_loop, NULL);
+}
+
+static void
+max_conns_message_complete (SoupSession *session, SoupMessage *msg, gpointer user_data)
+{
+	if (++msgs_done == TEST_CONNS)
+		g_main_loop_quit (max_conns_loop);
+}
+
+static void
+do_max_conns_test_for_session (SoupSession *session)
+{
+	SoupMessage *msgs[TEST_CONNS];
+	int i;
+	guint timeout_id;
+
+	max_conns_loop = g_main_loop_new (NULL, TRUE);
+
+	g_mutex_lock (server_mutex);
+
+	g_signal_connect (session, "request-started",
+			  G_CALLBACK (max_conns_request_started), NULL);
+	msgs_done = 0;
+	for (i = 0; i < TEST_CONNS; i++) {
+		msgs[i] = soup_message_new_from_uri ("GET", base_uri);
+		g_object_ref (msgs[i]);
+		soup_session_queue_message (session, msgs[i],
+					    max_conns_message_complete, NULL);
+	}
+
+	g_main_loop_run (max_conns_loop);
+	if (msgs_done != MAX_CONNS) {
+		debug_printf (1, "  Queued %d connections out of max %d?",
+			      msgs_done, MAX_CONNS);
+		errors++;
+	}
+	g_signal_handlers_disconnect_by_func (session, max_conns_request_started, NULL);
+
+	msgs_done = 0;
+	g_idle_add (idle_start_server, NULL);
+	timeout_id = g_timeout_add (1000, quit_loop, NULL);
+	g_main_loop_run (max_conns_loop);
+
+	for (i = 0; i < TEST_CONNS; i++) {
+		if (!SOUP_STATUS_IS_SUCCESSFUL (msgs[i]->status_code)) {
+			debug_printf (1, "    Message %d failed? %d %s\n",
+				      i, msgs[i]->status_code,
+				      msgs[i]->reason_phrase ? msgs[i]->reason_phrase : "-");
+			errors++;
+		}
+	}
+
+	if (msgs_done != TEST_CONNS) {
+		/* Clean up so we don't get a spurious "Leaked
+		 * session" error.
+		 */
+		for (i = 0; i < TEST_CONNS; i++)
+			soup_session_cancel_message (session, msgs[i], SOUP_STATUS_CANCELLED);
+		g_main_loop_run (max_conns_loop);
+		g_source_remove (timeout_id);
+	}
+
+	g_main_loop_unref (max_conns_loop);
+
+	for (i = 0; i < TEST_CONNS; i++)
+		g_object_unref (msgs[i]);
+}
+
+static void
+do_max_conns_test (void)
+{
+	SoupSession *session;
+
+	debug_printf (1, "\nExceeding max-conns\n");
+
+	debug_printf (1, "  Async session\n");
+	session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC,
+					 SOUP_SESSION_MAX_CONNS, MAX_CONNS,
+					 NULL);
+	do_max_conns_test_for_session (session);
+	soup_test_session_abort_unref (session);
+
+	debug_printf (1, "  Sync session\n");
+	session = soup_test_session_new (SOUP_TYPE_SESSION_SYNC,
+					 SOUP_SESSION_MAX_CONNS, MAX_CONNS,
+					 NULL);
+	do_max_conns_test_for_session (session);
+	soup_test_session_abort_unref (session);
+}
+
+static gboolean
+cancel_message_timeout (gpointer msg)
+{
+	SoupSession *session = g_object_get_data (G_OBJECT (msg), "session");
+
+	soup_session_cancel_message (session, msg, SOUP_STATUS_CANCELLED);
+	g_object_unref (msg);
+	g_object_unref (session);
+	return FALSE;
+}
+
+static gpointer
+cancel_message_thread (gpointer msg)
+{
+	SoupSession *session = g_object_get_data (G_OBJECT (msg), "session");
+
+	g_usleep (100000); /* .1s */
+	soup_session_cancel_message (session, msg, SOUP_STATUS_CANCELLED);
+	g_object_unref (msg);
+	g_object_unref (session);
+	return NULL;
+}
+
+static void
+do_cancel_while_reading_test_for_session (SoupSession *session)
+{
+	SoupMessage *msg;
+	GThread *thread = NULL;
+	SoupURI *uri;
+
+	uri = soup_uri_new_with_base (base_uri, "/slow");
+	msg = soup_message_new_from_uri ("GET", uri);
+	soup_uri_free (uri);
+
+	g_object_set_data (G_OBJECT (msg), "session", session);
+	g_object_ref (msg);
+	g_object_ref (session);
+	if (SOUP_IS_SESSION_ASYNC (session))
+		g_timeout_add (100, cancel_message_timeout, msg);
+	else
+		thread = g_thread_create (cancel_message_thread, msg, TRUE, NULL);
+
+	soup_session_send_message (session, msg);
+
+	if (msg->status_code != SOUP_STATUS_CANCELLED) {
+		debug_printf (1, "      FAILED: %d %s (expected Cancelled)\n",
+			      msg->status_code, msg->reason_phrase);
+		errors++;
+	}
+	g_object_unref (msg);
+
+	if (thread)
+		g_thread_join (thread);
+}
+
+static void
+do_cancel_while_reading_test (void)
+{
+	SoupSession *session;
+
+	debug_printf (1, "\nCancelling message while reading response\n");
+
+	debug_printf (1, "  Async session\n");
+	session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC, NULL);
+	do_cancel_while_reading_test_for_session (session);
+	soup_test_session_abort_unref (session);
+
+	debug_printf (1, "  Sync session\n");
+	session = soup_test_session_new (SOUP_TYPE_SESSION_SYNC, NULL);
+	do_cancel_while_reading_test_for_session (session);
+	soup_test_session_abort_unref (session);
+}
+
 int
 main (int argc, char **argv)
 {
@@ -758,6 +1013,8 @@ main (int argc, char **argv)
 
 	test_init (argc, argv, NULL);
 
+	server_mutex = g_mutex_new ();
+
 	server = soup_test_server_new (TRUE);
 	soup_server_add_handler (server, NULL, server_callback, NULL, NULL);
 	base_uri = soup_uri_new ("http://127.0.0.1/");
@@ -779,9 +1036,12 @@ main (int argc, char **argv)
 	do_content_length_framing_test ();
 	do_accept_language_test ();
 	do_persistent_connection_timeout_test ();
+	do_max_conns_test ();
+	do_cancel_while_reading_test ();
 
 	soup_uri_free (base_uri);
 	soup_test_server_quit_unref (server);
+	g_mutex_free (server_mutex);
 
 	test_cleanup ();
 	return errors != 0;
