@@ -81,7 +81,6 @@ test_init (int argc, char **argv, GOptionEntry *entries)
 	GError *error = NULL;
 	GTlsBackend *tls_backend;
 
-	g_thread_init (NULL);
 	g_type_init ();
 
 	name = strrchr (argv[0], '/');
@@ -230,7 +229,6 @@ soup_test_session_new (GType type, ...)
 
 	g_object_set (G_OBJECT (session),
 		      SOUP_SESSION_SSL_CA_FILE, SRCDIR "/test-cert.pem",
-		      SOUP_SESSION_SSL_STRICT, FALSE,
 		      NULL);
 
 	if (http_debug_level && !logger) {
